==> rlwrap_0.24.orig.tar.gz: rlwrap-0.24/src/readline.c <==
/*  readline.c: interacting with the GNU readline library */

/*  This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License , or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    You may contact the author by:
       e-mail:  hlub@knoware.nl
*/


#include "rlwrap.h"

/* global vars */
int remember_for_completion = FALSE;	/* whether we should put al words from in/output on the list */
char *multiline_separator = NULL;	/* character sequence to use in lieu of newline when storing multi-line input in a single history line */
struct rl_state saved_rl_state = { "", "", 0, 0 };	/* saved state of readline */
static char return_key;		/* Key pressed to enter line */
static int forget_line;


char previous_line[BUFFSIZE] = "";	/* previous input line */

/* forward declarations */
void line_handler(char *);
void my_add_history(char *);
int my_accept_line(int, int);
int my_accept_line_and_forget(int, int);
int munge_line_in_editor(int, int);

void
init_readline(char *prompt)
{
  rl_add_defun("rlwrap_accept_line", my_accept_line, -1);
  rl_bind_key('\n', my_accept_line);
  rl_bind_key('\r', my_accept_line);

  rl_add_defun("insert_close", rl_insert_close, -1);
  rl_bind_key(')', rl_insert_close);
  rl_bind_key('}', rl_insert_close);
  rl_bind_key(']', rl_insert_close);

  rl_add_defun("rlwrap_accept_line_and_forget", my_accept_line_and_forget,
	       -1);
  rl_bind_key(15, my_accept_line_and_forget);	/* ascii #15 (Control-O) is unused in readline's emacs and vi keymaps */

  if (multiline_separator != NULL) {
    rl_add_defun("rlwrap_call_editor", munge_line_in_editor, -1);
    rl_bind_key(30, munge_line_in_editor);	/* ascii #30 (Control-^) is unused in readline's emacs and vi keymaps */
  }

  rl_initialize();		/* This has to happen *before* we set rl_redisplay_function, otherwise
				   readline will not bother to call tgetent(), will be agnostic about terminal
				   capabilities and hence not be able to honour e.g. a set horizontal-scroll-mode off
				   in .inputrc */
  using_history();
  rl_redisplay_function = my_redisplay;
  rl_already_prompted = 1;	/* we'll always write our own prompt */
  rl_completion_entry_function =
    (rl_compentry_func_t *) & my_completion_function;
  rl_catch_signals = FALSE;
  rl_catch_sigwinch = FALSE;
}


void
save_rl_state(struct rl_state *state)
{

  mystrlcpy(state->text, rl_line_buffer, sizeof(state->text));	/* save text */
  mystrlcpy(state->prompt, rl_prompt, sizeof(state->prompt));
  state->point = rl_point;	/* and point    */
  rl_line_buffer[0] = '\0';
  if (state->already_saved)
    return;
  state->already_saved = 1;
  rl_delete_text(0, rl_end);	/* clear line  (after prompt) */
  rl_point = 0;
  my_redisplay();		/* and redisplay */
  rl_callback_handler_remove();	/* restore old term settings */
  rl_deprep_terminal();

}


void
restore_rl_state(struct rl_state *state)
{
  char *newprompt;

  newprompt = state->prompt;

  DPRINTF1(DEBUG_READLINE, "restore prompt: '%s'", newprompt);
  rl_expand_prompt(newprompt);	/* We have to call this because rl_already_prompted is set */

  mirror_slaves_echo_mode();	/* don't show passwords etc */
  rl_callback_handler_install(newprompt, &line_handler);
  rl_insert_text(state->text);
  rl_point = state->point;
  state->already_saved = 0;
  cr();
  rl_forced_update_display();
  rl_prep_terminal(1);
}

void
line_handler(char *line)
{
  char *rewritten_line;

  if (line == NULL) {		/* EOF, forward it  */
    DPRINTF1(DEBUG_READLINE, "EOF detected, writing character %d", term_eof);
    write_EOF_to_master_pty();
  } else {
    if (*line && strlen(line) > 1 && redisplay && !forget_line) {	/* don't save empty lines,  single keystrokes,
									   passwords, or lines entered with CTRL-F */
      my_add_history(line);
    }
    forget_line = 0;
    clear_line();
    write(STDOUT_FILENO, rl_prompt, strlen(rl_prompt));
    rewritten_line =
      (multiline_separator ?
       search_and_replace(multiline_separator, "\n", line, 0, NULL,
			  NULL) : mysavestring(line));

    put_in_output_queue(rewritten_line);
    DPRINTF2(DEBUG_READLINE, "writing %d bytes %s", strlen(rewritten_line),
	     mangle_string_for_debug_log(rewritten_line, 40));
    write_EOL_to_master_pty(return_key ? &return_key : "\n");
    free(line);
    free(rewritten_line);	/* we're done with them  */

    return_key = 0;
    within_line_edit = FALSE;

    rl_callback_handler_remove();
    set_echo(FALSE);
    saved_rl_state.text[0] = '\0';
    saved_rl_state.prompt[0] = '\0';
    saved_rl_state.point = 0;
    saved_rl_state.already_saved = TRUE;

  }
}




/* this function (drop-in replacement for readline's own accept-line())
   will be bound to RETURN key: */
int
my_accept_line(int count, int key)
{
  rl_point = 0;			/* leave cursor on predictable place */
  my_redisplay();
  rl_done = 1;
  return_key = (char)key;
  return 0;
}

int
my_accept_line_and_forget(int count, int key)
{
  forget_line = 1;
  return my_accept_line(count, '\n');
}


void
my_add_history(char *line)
{				/* add line to history list, avoiding duplicates */
  if (strncmp(line, previous_line, sizeof(previous_line))) {
    mystrlcpy(previous_line, line, sizeof(previous_line));
    add_history(line);
  }
}




/* Homegrown redisplay function - erases current line and prints the
   new one.  Used for passwords (where we want to show **** instead of
   user input) and whenever HOMEGROWN_REDISPLAY is defined (for
   systems where rl_redisplay() misbehaves, like sometimes on
   Solaris). Otherwise we use the much faster and smoother
   rl_redisplay() This function cannot display multiple lines: it will
   only scroll horizontally (even if horizontal-scroll-mode is off in
   .inputrc)
*/



static void
my_homegrown_redisplay(int hide_passwords)
{
  static int line_start = 0;	/* at which position of prompt_plus_line does the printed line start? */
  static int line_extends_right = 0;
  static int line_extends_left = 0;
  static char *previous_line = NULL;
  static int previous_curpos = 0;

  int width = window_size.ws_col;
  int skip = max(1, min(width / 5, 10));	/* jumpscroll this many positions when cursor reaches edge of terminal */
  int promptlen = strlen(rl_prompt);
  char *prompt_plus_line = add2strings(rl_prompt, rl_line_buffer);
  char *new_line;
  int total_length = strlen(prompt_plus_line);
  int curpos = promptlen + rl_point;
  int i, printed_length, new_curpos, keep_old_line;

  if (hide_passwords)
    for (i = promptlen; i < total_length; i++)
      prompt_plus_line[i] = '*';	/* make user input unreadable  */


  if (rl_point == 0)		/* re)set  at program start and after accept_line (where rl_point is zeroed) */
    line_start = 0;
  if (curpos - line_start > width - line_extends_right)	/* cursor falls off right edge ?   */
    line_start = (curpos - width + line_extends_right) + skip;	/* jumpscroll left                 */

  if (curpos < line_start + line_extends_left)	/* cursor falls off left edge ?    */
    if (curpos == total_length)	/* .. but still at end of line?    */
      line_start = max(0, total_length - width);	/* .. try to display entire line   */
    else			/* in not at end of line ..        */
      line_start = curpos - line_extends_left - skip;	/* ... jumpscroll right ..         */
  if (line_start < 0)
    line_start = 0;		/* ... but not past start of line! */

  printed_length = min(width, total_length - line_start);	/* never print more than width     */



  /* some invariants :     0 <= line_start <= curpos <= line_start + printed_length <= total_length */
  /* these are interesting:   ^                                                      ^              */

  assert(0 <= line_start);
  assert(line_start <= curpos);
  assert(curpos <= line_start + printed_length);	/* <=, rather than <, as cursor may be past eol   */
  assert(line_start + printed_length <= total_length);


  new_line = prompt_plus_line + line_start;
  new_line[printed_length] = '\0';
  new_curpos = curpos - line_start;

  /* indicate whether line extends past right or left edge  (i.e. whether the "interesting
     inequalities marked ^ above are really unequal) */

  line_extends_left = (line_start > 0 ? 1 : 0);
  line_extends_right = (total_length - line_start > width ? 1 : 0);
  if (line_extends_left)
    new_line[0] = '<';
  if (line_extends_right)
    new_line[width - 1] = '>';



  keep_old_line = FALSE;
  if (term_cursor_hpos) {
    if (previous_line && strcmp(new_line, previous_line) == 0) {
      keep_old_line = TRUE;
    } else {
      if (previous_line)
	free(previous_line);
      previous_line = mysavestring(new_line);
    }
  }


  if (!keep_old_line) {
    clear_line();
    cr();
    write(STDOUT_FILENO, new_line, printed_length);
  }
  free(prompt_plus_line);
  assert(term_cursor_hpos || !keep_old_line);	/* if we cannot position cursor, we must have reprinted ... */

  if (term_cursor_hpos)
    cursor_hpos(new_curpos);
  else				/* ... so we know we're 1 past last position on line */
    backspace(printed_length - new_curpos);
}



void
my_redisplay()
{
#ifndef HOMEGROWN_REDISPLAY
  if (redisplay)
    rl_redisplay();
  else
#endif
    my_homegrown_redisplay(!redisplay);
}




int
munge_line_in_editor(int count, int key)
{
  int line_number = 0, column_number = 0, tmpfile_OK, ret, tmpfile_fd, bytes_read;
  size_t tmpfilesize;
  char command[BUFFSIZE], *p, *tmpdir, *tmpfilename, *text_to_edit;
  char *editor_command1, *editor_command2, *editor_command3, *editor_command4,
    *line_number_as_string, *column_number_as_string;
  char *input, *rewritten_input, *rewritten_input2;
  char *possible_tmpdirs[] =
    { getenv("TMPDIR"), getenv("TMP"), getenv("TEMP"), "/tmp" };
  char *possible_editor_commands[] =
    { getenv("RLWRAP_EDITOR"), getenv("EDITOR"), getenv("VISUAL"), "vi +%L" };

#ifndef HAVE_SYSTEM
  return 0;			/* just being lazy */
#else

  /* create temporary filename */
#ifdef HAVE_MKSTEMP
  tmpdir = first_of(possible_tmpdirs);
  tmpfilename = add3strings(tmpdir, "/rlwrap_", "XXXXXX");
  tmpfile_OK = mkstemp(tmpfilename);
#else
  tmpfilename = mymalloc(L_tmpnam);
  tmpfile_OK = (int)tmpnam(tmpfilename);
#endif
  if (!tmpfile_OK)
    myerror("could not find unique temporary file name");

  /* write current input to it, replacing the newline substitute (multiline_separator) with the real thing */
  tmpfile_fd = open(tmpfilename, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR);
  if (tmpfile_fd < 0)
    myerror("could not create temporary file %s", tmpfilename);
  text_to_edit =
    search_and_replace(multiline_separator, "\n", rl_line_buffer, rl_point,
		       &line_number, &column_number);
  if (write(tmpfile_fd, text_to_edit, strlen(text_to_edit)) < 0)
    myerror("could not write to temporary file %s", tmpfilename);

  if (close(tmpfile_fd) != 0)
    myerror("couldn't close temporary file %s", tmpfilename);

  /* find out which editor command we have to use */

  editor_command1 = first_of(possible_editor_commands);
  line_number_as_string = as_string(line_number);
  column_number_as_string = as_string(column_number);
  editor_command2 =
    search_and_replace("%L", line_number_as_string, editor_command1, 0, NULL,
		       NULL);
  editor_command3 =
    search_and_replace("%C", column_number_as_string, editor_command2, 0,
		       NULL, NULL);
  editor_command4 = add3strings(editor_command3, " ", tmpfilename);

  /* call editor, temporarily restoring terminal settings */
  clear_line();
  cr();
  if (terminal_settings_saved && (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings) < 0))	/* reset terminal */
    myerror("tcsetattr error on stdin");
  DPRINTF1(DEBUG_READLINE, "calling %s", editor_command4);
  if (ret = system(editor_command4)) {
    if (WIFSIGNALED(ret)) {
      fprintf(stderr, "\n"); errno = 0;
      myerror("editor killed by signal");
    } else {    
      myerror("failed to invoke editor with '%s'", editor_command4);
    }
  }
  completely_mirror_slaves_terminal_settings();
  ignore_queued_input = TRUE;  

  /* read back edited input, replacing real newline with substitute */
  tmpfile_fd = open(tmpfilename, O_RDONLY);
  if (tmpfile_fd < 0)
    myerror("could not read temp file %s", tmpfilename);
  tmpfilesize = filesize(tmpfilename);
  input = mymalloc(tmpfilesize + 1);
  bytes_read = read(tmpfile_fd, input, tmpfilesize);
  if (bytes_read < 0)
    myerror("unreadable temp file %s", tmpfilename);
  input[bytes_read] = '\0';
  rewritten_input = search_and_replace("\t", "    ", input, 0, NULL, NULL);	/* rlwrap cannot handle tabs in input lines */
  rewritten_input2 =
    search_and_replace("\n", multiline_separator, rewritten_input, 0, NULL,
		       NULL);
  for(p=rewritten_input2; *p ;p++)
    if(*p < ' ')
      *p = ' ';        /* replace all control characters (like \r) by spaces */
  
  rl_delete_text(0, strlen(rl_line_buffer));
  rl_point = 0;
  rl_insert_text(rewritten_input2);
  rl_point = 0;			/* leave cursor on predictable place */
  rl_done = 1;
  return_key = (char)'\n';

  

  /* wash those dishes */
  if (unlink(tmpfilename))
    myerror("could not delete temporary file %s", tmpfilename);
  free(editor_command2);
  free(editor_command3);
  free(editor_command4);
  free(line_number_as_string);
  free(column_number_as_string);
  free(tmpfilename);
  free(text_to_edit);
  free(input);
  free(rewritten_input);
  free(rewritten_input2);

  return 0;
#endif /* HAVE_SYSTEM */
}

==> rlwrap_0.28.orig.tar.gz: rlwrap-0.28/src/readline.c <==
/*  readline.c: interacting with the GNU readline library */

/*  This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License , or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    You may contact the author by:
       e-mail:  hlub@knoware.nl
*/


#include "rlwrap.h"

/* global vars */
int remember_for_completion = FALSE;	/* whether we should put al words from in/output on the list */
char *multiline_separator = NULL;	/* character sequence to use in lieu of newline when storing multi-line input in a single history line */
char *pre_given = NULL;                 /* pre-given user input when rlwrap starts up */
struct rl_state saved_rl_state = { "", "", 0, 0 };	/* saved state of readline */
static char return_key;	                /* Key pressed to enter line */
static int forget_line;



/* forward declarations */
void line_handler(char *);
void my_add_history(char *);
int my_accept_line(int, int);
int my_accept_line_and_forget(int, int);
int munge_line_in_editor(int, int);

void
init_readline(char *prompt)
{
  rl_add_defun("rlwrap_accept_line", my_accept_line, -1);
  rl_bind_key('\n', my_accept_line);
  rl_bind_key('\r', my_accept_line);

  rl_add_defun("insert_close", rl_insert_close, -1);
  rl_bind_key(')', rl_insert_close);
  rl_bind_key('}', rl_insert_close);
  rl_bind_key(']', rl_insert_close);

  rl_add_defun("rlwrap_accept_line_and_forget", my_accept_line_and_forget,
	       -1);
  rl_bind_key(15, my_accept_line_and_forget);	/* ascii #15 (Control-O) is unused in readline's emacs and vi keymaps */

  if (multiline_separator != NULL) {
    rl_add_defun("rlwrap_call_editor", munge_line_in_editor, -1);
    rl_bind_key(30, munge_line_in_editor);	/* ascii #30 (Control-^) is unused in readline's emacs and vi keymaps */
  }
  /* rl_variable_bind("gnah","gnerp"); It is not possible to create new readline variables (only functions) */
  rl_initialize();		/* This has to happen *before* we set rl_redisplay_function, otherwise
				   readline will not bother to call tgetent(), will be agnostic about terminal
				   capabilities and hence not be able to honour e.g. a set horizontal-scroll-mode off
				   in .inputrc */
  /* printf("gnah = %s\n", rl_variable_value("gnah")); */ 
  using_history();
  rl_redisplay_function = my_redisplay;
  rl_already_prompted = 1;	/* we'll always write our own prompt */
  rl_completion_entry_function =
    (rl_compentry_func_t *) & my_completion_function;
  rl_catch_signals = FALSE;
  rl_catch_sigwinch = FALSE;
  if (pre_given) {
    mystrlcpy(saved_rl_state.text, pre_given, sizeof(saved_rl_state.text));
    saved_rl_state.point =  strlen(pre_given);
    DPRINTF0(DEBUG_READLINE, "Starting line edit (because of -P option)");
    within_line_edit = TRUE;
    restore_rl_state(&saved_rl_state);
  }	
}


void
save_rl_state(struct rl_state *state)
{

  mystrlcpy(state->text, rl_line_buffer, sizeof(state->text));	/* save text */
  mystrlcpy(state->prompt, rl_prompt, sizeof(state->prompt));
  state->point = rl_point;	/* and point    */
  rl_line_buffer[0] = '\0';
  if (state->already_saved)
    return;
  state->already_saved = 1;
  rl_delete_text(0, rl_end);	/* clear line  (after prompt) */
  rl_point = 0;
  my_redisplay();		/* and redisplay */
  rl_callback_handler_remove();	/* restore old term settings */
  rl_deprep_terminal();

}


void
restore_rl_state(struct rl_state *state)
{
  char *newprompt;

  newprompt = state->prompt;

  DPRINTF1(DEBUG_READLINE, "restore prompt: '%s'", newprompt);
  rl_expand_prompt(newprompt);	/* We have to call this because rl_already_prompted is set */

  mirror_slaves_echo_mode();	/* don't show passwords etc */
  rl_callback_handler_install(newprompt, &line_handler);
  rl_insert_text(state->text);
  rl_point = state->point;
  state->already_saved = 0;
  cr();
  rl_forced_update_display();
  rl_prep_terminal(1);
}

void
line_handler(char *line)
{
  char *rewritten_line;

  if (line == NULL) {		/* EOF, forward it  */
    DPRINTF1(DEBUG_READLINE, "EOF detected, writing character %d", term_eof);
    write_EOF_to_master_pty();
  } else {
    if (*line && redisplay && !forget_line) {	/* don't save empty lines,  single keystrokes,
						   passwords, or lines entered with CTRL-F */
      my_add_history(line);
    }
    forget_line = 0;
    if (rl_editing_mode == 0) { 	/* vi mode */
      curs_up();                        /* @@@ kludge (yet another one...) : when a line is accepted in vi mode,
					   the cursor is already on the next line . We put it back up */
    }	

    clear_line();
    write(STDOUT_FILENO, rl_prompt, strlen(rl_prompt));
    rewritten_line =
      (multiline_separator ?
       search_and_replace(multiline_separator, "\n", line, 0, NULL,
			  NULL) : mysavestring(line));

    put_in_output_queue(rewritten_line);
    DPRINTF2(DEBUG_READLINE, "writing %d bytes %s", strlen(rewritten_line),
	     mangle_string_for_debug_log(rewritten_line, 40));
    write_EOL_to_master_pty(return_key ? &return_key : "\n");
    /* if (one_shot_rlwrap)
       write_EOF_to_master_pty(); */
    free(line);
    free(rewritten_line);	/* we're done with them  */

    return_key = 0;
    within_line_edit = FALSE;

    rl_callback_handler_remove();
    set_echo(FALSE);
    saved_rl_state.text[0] = '\0';
    saved_rl_state.prompt[0] = '\0';
    saved_rl_state.point = 0;
    saved_rl_state.already_saved = TRUE;
  }
}




/* this function (drop-in replacement for readline's own accept-line())
   will be bound to RETURN key: */
int
my_accept_line(int count, int key)
{
  rl_point = 0;			/* leave cursor on predictable place */
  my_redisplay();
  rl_done = 1;
  return_key = (char)key;
  return 0;
}

int
my_accept_line_and_forget(int count, int key)
{
  forget_line = 1;
  return my_accept_line(count, '\n');
}


/* format line and add it to history list, avoiding duplicates if necessary */
void
my_add_history(char *line)
{	
  int lookback, count, here;
  char *new_entry;

  
  switch (history_duplicate_avoidance_policy) { 
  case KEEP_ALL_DOUBLES:
    lookback = 0; break;
  case ELIMINATE_SUCCESIVE_DOUBLES:
    lookback = 1; break;
  case ELIMINATE_ALL_DOUBLES:
    lookback = history_length; break;
  }

  if (!history_format)  /* enter it as-is */
    new_entry = mysavestring(line);	
  else 
    new_entry = append_and_expand_history_format(line);

   
  lookback = min(history_length, lookback);      
  for (count = 0, here = history_length + history_base - 1;
       count < lookback ;
       count++, here--) {
    /* DPRINTF3(DEBUG_READLINE, "strcmping <%s> and <%s> (count = %d)", line, history_get(here)->line ,count); */
    if (strncmp(new_entry, history_get(here) -> line, 10000) == 0) {
      HIST_ENTRY *entry = remove_history (here - history_base); /* according to the history library doc this should be
								   remove_history(here) @@@?*/
      DPRINTF2(DEBUG_READLINE, "removing duplicate entry #%d (%s)", here, entry->line);
      free(entry->line);
      free(entry);
    }
  }
  add_history(new_entry);
  free(new_entry);
}




/* Homegrown redisplay function - erases current line and prints the
   new one.  Used for passwords (where we want to show **** instead of
   user input) and whenever HOMEGROWN_REDISPLAY is defined (for
   systems where rl_redisplay() misbehaves, like sometimes on
   Solaris). Otherwise we use the much faster and smoother
   rl_redisplay() This function cannot display multiple lines: it will
   only scroll horizontally (even if horizontal-scroll-mode is off in
   .inputrc)
*/



static void
my_homegrown_redisplay(int hide_passwords)
{
  static int line_start = 0;	/* at which position of prompt_plus_line does the printed line start? */
  static int line_extends_right = 0;
  static int line_extends_left = 0;
  static char *previous_line = NULL;
  static int previous_curpos = 0;

  int width = window_size.ws_col;
  int skip = max(1, min(width / 5, 10));	/* jumpscroll this many positions when cursor reaches edge of terminal */
  int promptlen = strlen(rl_prompt);
  char *prompt_plus_line = add2strings(rl_prompt, rl_line_buffer);
  char *new_line;
  int total_length = strlen(prompt_plus_line);
  int curpos = promptlen + rl_point;
  int i, printed_length, new_curpos, keep_old_line;

  if (hide_passwords)
    for (i = promptlen; i < total_length; i++)
      prompt_plus_line[i] = '*';	/* make user input unreadable  */


  if (rl_point == 0)		/* re)set  at program start and after accept_line (where rl_point is zeroed) */
    line_start = 0;
  if (curpos - line_start > width - line_extends_right)	/* cursor falls off right edge ?   */
    line_start = (curpos - width + line_extends_right) + skip;	/* jumpscroll left                 */

  if (curpos < line_start + line_extends_left)	/* cursor falls off left edge ?    */
    if (curpos == total_length)	/* .. but still at end of line?    */
      line_start = max(0, total_length - width);	/* .. try to display entire line   */
    else			/* in not at end of line ..        */
      line_start = curpos - line_extends_left - skip;	/* ... jumpscroll right ..         */
  if (line_start < 0)
    line_start = 0;		/* ... but not past start of line! */

  printed_length = min(width, total_length - line_start);	/* never print more than width     */



  /* some invariants :     0 <= line_start <= curpos <= line_start + printed_length <= total_length */
  /* these are interesting:   ^                                                      ^              */

  assert(0 <= line_start);
  assert(line_start <= curpos);
  assert(curpos <= line_start + printed_length);	/* <=, rather than <, as cursor may be past eol   */
  assert(line_start + printed_length <= total_length);


  new_line = prompt_plus_line + line_start;
  new_line[printed_length] = '\0';
  new_curpos = curpos - line_start;

  /* indicate whether line extends past right or left edge  (i.e. whether the "interesting
     inequalities marked ^ above are really unequal) */

  line_extends_left = (line_start > 0 ? 1 : 0);
  line_extends_right = (total_length - line_start > width ? 1 : 0);
  if (line_extends_left)
    new_line[0] = '<';
  if (line_extends_right)
    new_line[width - 1] = '>';



  keep_old_line = FALSE;
  if (term_cursor_hpos) {
    if (previous_line && strcmp(new_line, previous_line) == 0) {
      keep_old_line = TRUE;
    } else {
      if (previous_line)
	free(previous_line);
      previous_line = mysavestring(new_line);
    }
  }


  if (!keep_old_line) {
    clear_line();
    cr();
    write(STDOUT_FILENO, new_line, printed_length);
  }
  free(prompt_plus_line);
  assert(term_cursor_hpos || !keep_old_line);	/* if we cannot position cursor, we must have reprinted ... */

  if (term_cursor_hpos)
    cursor_hpos(new_curpos);
  else				/* ... so we know we're 1 past last position on line */
    backspace(printed_length - new_curpos);
}



void
my_redisplay()
{
#ifndef HOMEGROWN_REDISPLAY
  if (redisplay)
    rl_redisplay();
  else
#endif
    my_homegrown_redisplay(!redisplay);
}




int
munge_line_in_editor(int count, int key)
{
  int line_number = 0, column_number = 0, tmpfile_OK, ret, tmpfile_fd, bytes_read;
  size_t tmpfilesize;
  char command[BUFFSIZE], *p, *tmpdir, *tmpfilename, *text_to_edit;
  char *editor_command1, *editor_command2, *editor_command3, *editor_command4,
    *line_number_as_string, *column_number_as_string;
  char *input, *rewritten_input, *rewritten_input2;
  char *possible_tmpdirs[] =
    { getenv("TMPDIR"), getenv("TMP"), getenv("TEMP"), "/tmp" };
  char *possible_editor_commands[] =
    { getenv("RLWRAP_EDITOR"), getenv("EDITOR"), getenv("VISUAL"), "vi +%L" };

#ifndef HAVE_SYSTEM
  return 0;    /* we could write our own system() using execv(), but we are too lazy to do that */
#else

  /* create temporary filename */
#ifdef HAVE_MKSTEMP
  tmpdir = first_of(possible_tmpdirs);
  tmpfilename = add3strings(tmpdir, "/rlwrap_", "XXXXXX");
  tmpfile_OK = mkstemp(tmpfilename);
#else
  tmpfilename = mymalloc(L_tmpnam);
  tmpfile_OK = (int)tmpnam(tmpfilename);
#endif
  if (!tmpfile_OK)
    myerror("could not find unique temporary file name");

  /* write current input to it, replacing the newline substitute (multiline_separator) with the real thing */
  tmpfile_fd = open(tmpfilename, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR);
  if (tmpfile_fd < 0)
    myerror("could not create temporary file %s", tmpfilename);
  text_to_edit =
    search_and_replace(multiline_separator, "\n", rl_line_buffer, rl_point,
		       &line_number, &column_number);
  if (write(tmpfile_fd, text_to_edit, strlen(text_to_edit)) < 0)
    myerror("could not write to temporary file %s", tmpfilename);

  if (close(tmpfile_fd) != 0) /* improbable */
    myerror("couldn't close temporary file %s", tmpfilename); 

  /* find out which editor command we have to use */

  editor_command1 = first_of(possible_editor_commands);
  line_number_as_string = as_string(line_number);
  column_number_as_string = as_string(column_number);
  editor_command2 =
    search_and_replace("%L", line_number_as_string, editor_command1, 0, NULL,
		       NULL);
  editor_command3 =
    search_and_replace("%C", column_number_as_string, editor_command2, 0,
		       NULL, NULL);
  editor_command4 = add3strings(editor_command3, " ", tmpfilename);

  /* call editor, temporarily restoring terminal settings */
  clear_line();
  cr();
  if (terminal_settings_saved && (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings) < 0))	/* reset terminal */
    myerror("tcsetattr error on stdin");
  DPRINTF1(DEBUG_READLINE, "calling %s", editor_command4);
  if (ret = system(editor_command4)) {
    if (WIFSIGNALED(ret)) {
      fprintf(stderr, "\n"); errno = 0;
      myerror("editor killed by signal");
    } else {    
      myerror("failed to invoke editor with '%s'", editor_command4);
    }
  }
  completely_mirror_slaves_terminal_settings();
  ignore_queued_input = TRUE;  

  /* read back edited input, replacing real newline with substitute */
  tmpfile_fd = open(tmpfilename, O_RDONLY);
  if (tmpfile_fd < 0)
    myerror("could not read temp file %s", tmpfilename);
  tmpfilesize = filesize(tmpfilename);
  input = mymalloc(tmpfilesize + 1);
  bytes_read = read(tmpfile_fd, input, tmpfilesize);
  if (bytes_read < 0)
    myerror("unreadable temp file %s", tmpfilename);
  input[bytes_read] = '\0';
  rewritten_input = search_and_replace("\t", "    ", input, 0, NULL, NULL);	/* rlwrap cannot handle tabs in input lines */
  rewritten_input2 =
    search_and_replace("\n", multiline_separator, rewritten_input, 0, NULL,
		       NULL);
  for(p = rewritten_input2; *p ;p++)
    if(*p < ' ')
      *p = ' ';        /* replace all control characters (like \r) by spaces */
  
  rl_delete_text(0, strlen(rl_line_buffer));
  rl_point = 0;
  rl_insert_text(rewritten_input2);
  rl_point = 0;			/* leave cursor on predictable place */
  rl_done = 1;
  return_key = (char)'\n';

  

  /* wash those dishes */
  if (unlink(tmpfilename))
    myerror("could not delete temporary file %s", tmpfilename);
  free(editor_command2);
  free(editor_command3);
  free(editor_command4);
  free(line_number_as_string);
  free(column_number_as_string);
  free(tmpfilename);
  free(text_to_edit);
  free(input);
  free(rewritten_input);
  free(rewritten_input2);

  return 0;
#endif /* HAVE_SYSTEM */
}
